/*****************************************************************************/
//
//  variadic_helper.hpp: Helper classes/functions for variadic templates
//
//   $URL: https://yak.myhome.cx/repos/source/mylib/trunk/cpp/variadic_helper.hpp $
//  $Date: 2007-10-24 09:33:34 +0900 (Wed, 24 Oct 2007) $
//   $Rev: 475 $
//
//      Copyright (c) 2007 Yak!
//
//  This code is distributed under the zlib/libpng License.
//  See http://www.opensource.org/licenses/zlib-license.php
//
/*****************************************************************************/

#ifndef YAK_UTIL_VARIADIC_HELPER_HPP
#define YAK_UTIL_VARIADIC_HELPER_HPP

#include <boost/mpl/vector.hpp>
#include <boost/mpl/push_back.hpp>
#include <boost/mpl/at.hpp>
#include <boost/mpl/size.hpp>

#include <boost/preprocessor/iteration/local.hpp>
#include <boost/preprocessor/repetition/enum.hpp>

#ifdef YAK_USE_BOOST_TUPLE

#include <boost/tuple/tuple.hpp>

namespace yak {

namespace util {

namespace detail {

using boost::get;

} // namespace detail

// Convert variadic type parameters into a MPL vector

namespace detail {

template<typename Result, typename ... Tn>
struct variadic_to_vector_;

template<typename Result>
struct variadic_to_vector_<Result>
{
	typedef Result type;
};

template<typename Result, typename T0, typename ... Tn>
struct variadic_to_vector_<Result, T0, Tn...>
{
	typedef typename variadic_to_vector_<
		typename boost::mpl::push_back<Result, T0>::type,
		Tn...
	>::type type;
};

} // namespace detail

template<typename ... Tn>
struct variadic_to_vector
{
	typedef typename detail::variadic_to_vector_<boost::mpl::vector<>, Tn...>::type type;
};

// Convert variadic template type parameters to fixed-number type parameters

namespace detail {

// Result of push_back could be different type from boost::mpl::vector.
// Therefore, we do not use partial specialization with vector<...>

template<typename Lambda, typename VT, int N>
struct variadic_to_fixed_2;

// Partial specialization
// Expanded by Boost PP accroding to the following template
/* 
template<typename MetaFunc, typename VT>
struct variadic_to_fixed_2<MetaFunc, VT, n>
{
	typedef typename MetaFunc::template apply<
		typename boost::mpl::at_c<VT, 0>::type,
		...
		typename boost::mpl::at_c<VT, n-1>::type
	>::type type;
};
*/
// Alternative: typedef typename boost::mpl::apply<Lambda, ...>
// We do not use this alternative due to lmitiation of arity of metafunction.

template<typename MetaFunc, typename VT>
struct variadic_to_fixed_
{
	typedef typename variadic_to_fixed_2<MetaFunc, VT, boost::mpl::size<VT>::value>::type type;
};

} // namespace detail

template<typename MetaFunc, typename ... Tn>
struct variadic_to_fixed
{
	typedef typename detail::variadic_to_fixed_<MetaFunc, typename variadic_to_vector<Tn...>::type>::type type;
};


// variadic to tuple

namespace detail {

struct to_tuple
{
	template<typename T0 = boost::tuples::null_type,
	         typename T1 = boost::tuples::null_type,
	         typename T2 = boost::tuples::null_type,
	         typename T3 = boost::tuples::null_type,
	         typename T4 = boost::tuples::null_type,
	         typename T5 = boost::tuples::null_type,
	         typename T6 = boost::tuples::null_type,
	         typename T7 = boost::tuples::null_type,
	         typename T8 = boost::tuples::null_type,
	         typename T9 = boost::tuples::null_type
	>
	struct apply
	{
		typedef boost::tuple<T0, T1, T2, T3, T4, T5, T6, T7, T8, T9> type;
	};
};

} // namespace detail

template<typename ... Tn>
struct variadic_to_tuple
{
	typedef typename variadic_to_fixed<detail::to_tuple, Tn...>::type type;
};

// tuple to variadic

namespace detail {

template<int N, typename T, typename MetaFunc>
struct tuple_to_variadic_;

/*
template<typename T, typename MetaFunc>
struct tuple_to_variadic_<n, T, MetaFunc>
{
	typedef typename MetaFunc::template apply<
		typename boost::tuples::element<0, T>::type,
		...
		typename boost::tuples::element<n-1, T>::type
	>::type type;
};
*/

// TODO: result_of

template<int N, typename T, typename Func, typename R>
struct tuple_to_variadic_2;

/*
template<typename T, typename Func, typename R>
struct tuple_to_variadic_2<n, T, Func, R>
{
	R operator()(const T& t, Func f)
	{
		return f(boost::get<0>(t),
		  ...
		  boost::get<n-1>(t)
		);
	}
};
*/

} // namespace detail

// TODO: function and metafunction could be merged in the way simliar as Boost Fusion

template<typename T, typename MetaFunc>
struct tuple_to_variadic_type :
	public detail::tuple_to_variadic_<
		boost::tuples::length<T>::value,
		T,
		MetaFunc
	>::type
{
};

// TODO: result_of

template<typename R, typename T, typename Func>
R tuple_to_variadic(const T& t, Func f)
{
	return detail::tuple_to_variadic_2<boost::tuples::length<T>::value, T, Func, R>()(t, f);
}

// Boiler-plate specialization definition expanded by Boost PP

namespace detail {

#define TYPE(z, i, data) typename boost::mpl::at_c<VT, i>::type
#define ELEMENT_TYPE(z, i, data) typename boost::tuples::element<i, T>::type
#define GET(z, i, data) boost::get<i>(t)

#define BOOST_PP_LOCAL_MACRO(n) \
	template<typename MetaFunc, typename VT> \
	struct variadic_to_fixed_2<MetaFunc, VT, n> { \
		typedef typename MetaFunc::template apply< \
			BOOST_PP_ENUM(n, TYPE, _) \
		>::type type; \
	}; \
	template<typename T, typename MetaFunc> \
	struct tuple_to_variadic_<n, T, MetaFunc> \
	{ \
		typedef typename MetaFunc::template apply< \
			BOOST_PP_ENUM(n, ELEMENT_TYPE, _) \
		>::type type; \
	}; \
	template<typename T, typename Func, typename R> \
	struct tuple_to_variadic_2<n, T, Func, R> \
	{ \
		R operator()(const T& t, Func f) \
		{ \
			return f(BOOST_PP_ENUM(n, GET, _)); \
		} \
	};
#define BOOST_PP_LOCAL_LIMITS (0, 10)
#include BOOST_PP_LOCAL_ITERATE()

#undef TYPE
#undef ELEMENT
#undef GET
#undef BOOST_PP_LOCAL_MACRO
#undef BOOST_PP_LOCAL_LIMITS

} // namespace detail

} // namespace util

} // namespace yak

#else // YAK_USE_BOOST_TUPLE

#include <tuple>

namespace yak {

namespace util {

namespace detail {

using std::get;

} // namespace detail

// variadic to tuple

// just pass through variadic type parameters

template<typename ... Tn>
struct variadic_to_tuple
{
	typedef std::tuple<Tn...> type;
};

// tuple to variadic

namespace detail {

template<typename R, typename T, typename Func>
struct tuple_to_variadic_;

template<typename R, typename ... Tn, typename Func>
struct tuple_to_variadic_<R, std::tuple<Tn...>, Func>
{
private:
	R call(const std::tuple<Tn...>& t, Func f, const Tn& ... tn)
	{
		return f(tn...);
	}
	template<typename ... Un>
	R call(const std::tuple<Tn...>& t, Func f, const Un& ... un)
	{
		return call(t, f, un..., std::get<sizeof...(Un)>(t));
	}
public:
	R operator()(const std::tuple<Tn...>& t, Func f)
	{
		return call(t, f);
	}
};

} // namespace detail

template<typename R, typename T, typename Func>
R tuple_to_variadic(const T& t, Func f)
{
	return detail::tuple_to_variadic_<R, T, Func>()(t, f);
}

} // namespace util

} // namespace yak

#endif // YAK_USE_BOOST_TUPLE

#endif
